Unit Object3D;

{ 3D Objects & ZBuffer

Copyright (C) 1997-98 TothPaul@mygale.org
 http://www.mygale.org/~tothpaul

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}

Interface

{$N+}

Uses Tweak;

Type Float=Single;

{.$DEFINE FLOAT}

{$IFDEF FLOAT}
 Type Int=Float;
{$ELSE}
 Type Int=Integer;
{$ENDIF}

Type
 TVertice=record x,y,z:int end;
 TFace   =record a,b,c:word end;
 TVector =record x,y,z:int end;
 TCoord  =record x,y,z:int end;

 TVertices=array[0..1000] of TVertice;
 TFaces   =array[0..1000] of TFace;
 TVectors =array[0..1000] of TVector;
 TCoords  =array[0..1000] of TCoord;

 TObject3D=object
  VerticeCount:word;
  Vertices    :^TVertices;
  FaceCount   :word;
  Faces       :^TFaces;
  FNormals    :^TVectors;
  VNormals    :^TVectors;
  Coords      :^TCoords;
  Shades      :^TVectors;
  Hidden      :^TVectors;
  MnX,MnY,MnZ :int;
  MxX,MxY     :int;
  Constructor Create(AVertices,AFaces:word);
  Destructor Destroy;
  Procedure ComputeNormals(Normal:word);
  Procedure Translate(tx,ty,tz:int);
  Procedure Rotate(PhiX,PhiY,PhiZ:word);
  Procedure Project;
  Procedure DrawGouraud(AShade:byte);
  Procedure Draw(AShade:byte);
 end;

 TTorus=Object(TObject3D)
  Step1:word;
  Step2:word;
  Size :word;
  Constructor Create(AStep1,AStep2,Radius1,Radius2:word);
  Procedure SetSize(Radius1,Radius2:word);
 end;

 TCube=Object(TObject3D)
  Constructor Create(Width,Height,Depth:word);
 end;

 TCylinder=object(TObject3D)
  Constructor Create(Step,Height,Radius1,Radius2:word);
 end;

 TSphere=Object(TObject3D)
  Constructor Create(Step,Radius:word);
 end;

 TGSphere=Object(TObject3D)
  Constructor Create(Step,Radius:word);
 end;

Var
 Light:TVector;

{procedure InitTrigo(Phix,Phiy,Phiz:word);}

Implementation

{- Trigo ----------------------------------------------------------}
const
 FACTOR=1 SHL 8;

var
{$IFDEF FLOAT}
 sinx,siny,sinz:int;
 cosx,cosy,cosz:int;

Function iRound(i:int):integer;
 begin
  if i<-32768/2 then iRound:=-32768 div 2 else
  if i>+32767/2 then iRound:=+32767 div 2 else iRound:=Round(i);
 end;

procedure RotateFloat(Var S,D; count:word);
 Var
  src:array[0..1000] of TVertice absolute S;
  dst:array[0..1000] of TVertice absolute D;
  i:integer;
  rx,ry,rz:int;
 begin
  for i:=0 to count-1 do begin
   with src[i] do begin
   { rotate x : y'=y*cosx-z*sinx / z'=y*sinx+z*cosx}
    ry:= y*cosx- z*sinx;
    rz:= y*sinx+ z*cosx;
   { rotate y : x'=x*cosy+z*siny / z'=z*cosy-x*siny}
    rx:= x*cosy+rz*siny;
    rz:=rz*cosy- x*siny;
   end;
   with dst[i] do begin
   { rotate z : x'=x*cosz-y*sinz / y'=x*sinz+y*cosz }
     x:=rx*cosz-ry*sinz;
     y:=rx*sinz+ry*cosz;
     z:=rz;
   end;
  end;
 end;
{$ELSE}
 sintab,costab:array[0..255] of integer;
 sinx,siny,sinz:integer;
 cosx,cosy,cosz:integer;
 rx,ry,rz:word;

procedure RotateTrigo(src,dst:pointer;count:word); assembler;
 asm
 { rotate }

  mov si,word(src)
  mov di,word(dst)

  mov cx,count
  @next_point:

    { get point }
    mov es,word(src+2)
    mov ax,es:[si+0]; mov rx,ax
    mov ax,es:[si+2]; mov ry,ax
    mov ax,es:[si+4]; mov rz,ax

    { rotate x : y'=y*cosx-z*sinx / z'=y*sinx+z*cosx}
    mov ax,rz
    imul sinx
    mov bl,ah; mov bh,dl

    mov ax,ry
    imul cosx
    mov al,ah; mov ah,dl
    sub ax,bx
    xchg ax,ry

    imul sinx
    mov bl,ah; mov bh,dl
    mov ax,rz
    imul cosx
    mov al,ah; mov ah,dl
    add ax,bx
    mov rz,ax

    { rotate y : x'=x*cosy+z*siny / z'=z*cosy-x*siny}
    mov ax,rx
   imul cosy
    mov bl,ah; mov bh,dl
    mov ax,rz
   imul siny
    mov al,ah; mov ah,dl
    add ax,bx
    xchg rx,ax

   imul siny
    mov bl,ah; mov bh,dl
    mov ax,rz
   imul cosy
    mov al,ah; mov ah,dl
    sub ax,bx
    mov rz,ax

    { rotate z : x'=x*cosz-y*sinz / y'=x*sinz+y*cosz }
    mov ax,ry
   imul sinz
    mov bl,ah; mov bh,dl
    mov ax,rx
   imul cosz
    mov al,ah; mov ah,dl
    sub ax,bx
    xchg ax,rx

   imul sinz
    mov bl,ah; mov bh,dl
    mov ax,ry
   imul cosz
    mov al,ah; mov ah,dl
    add ax,bx
    mov ry,ax

    { set point }
    mov es,word(dst+2)
    mov ax,rx; mov es:[di+0],ax;
    mov ax,ry; mov es:[di+2],ax;
    mov ax,rz; mov es:[di+4],ax;

    add si,6
    add di,6

   dec cx
   jnz @next_point
 (*  loop @next_point *)
 end;
{$ENDIF}

{- Shade ----------------------------------------------------------}

function Shade(const vector:TVector):integer;
 begin
  Shade:={ABS}({$IFDEF FLOAT}iRound{$ENDIF}(
             (Light.x*Vector.x)
            +(Light.y*Vector.y)
            +(Light.z*Vector.z)
            ) div 100);
 end;

{- Object3D -------------------------------------------------------}

Constructor TObject3D.Create(AVertices,AFaces:word);
 begin
  VerticeCount:=AVertices;
  FaceCount:=AFaces;
  GetMem(Vertices,VerticeCount*SizeOf(TVertice));
  GetMem(Faces,FaceCount*SizeOf(TFace));
  GetMem(VNormals,VerticeCount*SizeOf(TVector));
  GetMem(FNormals,FaceCount*SizeOf(TVector));
  GetMem(Coords,VerticeCount*SizeOf(TCoord));
  GetMem(Shades,VerticeCount*SizeOf(TVector));
  GetMem(Hidden,FaceCount*SizeOf(TVector));
 end;

Destructor TObject3D.Destroy;
 begin
  FreeMem(Hidden,FaceCount*SizeOf(TVector));
  FreeMem(Shades,VerticeCount*SizeOf(TVector));
  FreeMem(Coords,VerticeCount*SizeOf(TCoord));
  FreeMem(FNormals,FaceCount*SizeOf(TVector));
  FreeMem(VNormals,VerticeCount*SizeOf(TVector));
  FreeMem(Faces,FaceCount*SizeOf(TFace));
  FreeMem(Vertices,VerticeCount*SizeOf(TVertice));
 end;

Procedure TObject3D.ComputeNormals(Normal:word);
 var
  i,j,nf:word;
  x1,y1,z1,x2,y2,z2,vl:single;
  nx,ny,nz:single;
 begin

  For i:=0 to FaceCount-1 do begin
   with Faces^[i] do begin
    x1:=Vertices^[b].x-Vertices^[a].x;
    y1:=Vertices^[b].y-Vertices^[a].y;
    z1:=Vertices^[b].z-Vertices^[a].z;
    x2:=Vertices^[c].x-Vertices^[a].x;
    y2:=Vertices^[c].y-Vertices^[a].y;
    z2:=Vertices^[c].z-Vertices^[a].z;
   end;
   nx:=y1*z2-y2*z1;
   ny:=z1*x2-z2*x1;
   nz:=x1*y2-x2*y1;
   vl:=sqrt(nx*nx+ny*ny+nz*nz);
   if vl=0 then vl:=1;
   with FNormals^[i] do begin
    x:={$IFNDEF FLOAT}Round{$ENDIF}((nx/vl)*Normal);
    y:={$IFNDEF FLOAT}Round{$ENDIF}((ny/vl)*Normal);
    z:={$IFNDEF FLOAT}Round{$ENDIF}((nz/vl)*Normal);
   end;
  end;

  for i:=0 to VerticeCount-1 do begin
   x1:=0; y1:=0; z1:=0; nf:=0;
   for j:=0 to FaceCount-1 do begin
    with Faces^[j] do if (a=i)or(b=i)or(c=i) then begin
     with FNormals^[j] do begin
      x1:=x1+x; y1:=y1+y; z1:=z1+z;
     end;
     inc(nf);
    end;
   end;
   if nf<>0 then begin
    x1:=x1/nf; y1:=y1/nf; z1:=z1/nf;
    vl:=sqrt(x1*x1+y1*y1+z1*z1);
    if vl=0 then vl:=1;
    with VNormals^[i] do begin
     x:={$IFNDEF FLOAT}Round{$ENDIF}((x1/vl)*Normal);
     y:={$IFNDEF FLOAT}Round{$ENDIF}((y1/vl)*Normal);
     z:={$IFNDEF FLOAT}Round{$ENDIF}((z1/vl)*Normal);
    end;
   end;
  end;

 end;

procedure TObject3D.Translate(tx,ty,tz:int);
{$IFDEF FLOAT}
 var
  i:integer;
 begin
  for i:=0 to VerticeCount-1 do with Vertices^[i] do begin
   Coords^[i].x:=x+tx;
   Coords^[i].y:=y+ty;
   Coords^[i].z:=z+tz;
  end;
 end;
{$ELSE}
 var
  _Count:word;
  _Coords:pointer;
  _Vertices:pointer;
 begin
  _Count   :=VerticeCount; { ES:[DI].Count }
  _Coords  :=Coords;       { ES:[DI].Coords }
  _Vertices:=Vertices;     { ES:[DI].Vertices }
  asm
  push ds
   les di,_Coords
   mov bx,tx
   mov cx,ty
   mov dx,tz
   lds si,_Vertices
   @next_point:

    lodsw; add ax,bx; stosw
    lodsw; add ax,cx; stosw
    lodsw; add ax,dx; stosw

   dec _Count
   jnz @next_point
  pop ds
  end;
 end;
{$ENDIF}

procedure TObject3D.Rotate(PhiX,PhiY,PhiZ:word);
{$IFDEF FLOAT}
 var
  i:integer;
 begin
  sinx:=Sin(PhiX*2*PI/255); cosx:=Cos(PhiX*2*PI/255);
  siny:=Sin(PhiY*2*PI/255); cosy:=Cos(PhiY*2*PI/255);
  sinz:=Sin(PhiZ*2*PI/255); cosz:=Cos(PhiZ*2*PI/255);
  RotateFloat(Coords^,Coords^,VerticeCount);
  RotateFloat(VNormals^,Shades^,VerticeCount);
  RotateFloat(FNormals^,Hidden^,FaceCount);
 end;
{$ELSE}
 begin
  asm
   mov si,Offset SinTab
   mov di,Offset CosTab

   mov bx,PhiX; and bx,255; add bx,bx
   mov ax,[bx+si]; mov sinx,ax
   mov ax,[bx+di]; mov cosx,ax

   mov bx,PhiY; and bx,255; add bx,bx
   mov ax,[bx+si]; mov siny,ax
   mov ax,[bx+di]; mov cosy,ax

   mov bx,PhiZ; and bx,255; add bx,bx
   mov ax,[bx+si]; mov sinz,ax
   mov ax,[bx+di]; mov cosz,ax
  end;

  RotateTrigo(Coords,Coords,VerticeCount);
  RotateTrigo(VNormals,Shades,VerticeCount);
  RotateTrigo(FNormals,Hidden,FaceCount);
 end;
{$ENDIF}

procedure TObject3D.Project;
{$IFDEF FLOAT}
 var
  i:integer;
 begin
  MnX:=MaxX; MxX:=-1;
  MnY:=MaxY; MxY:=-1; MnZ:=CENTER_Z;
  for i:=0 to VerticeCount-1 do begin
   with Coords^[i] do begin
    x:=CENTER_X+x/(z+CENTER_Z)*FACTOR;
    y:=CENTER_Y-y/(z+CENTER_Z)*FACTOR;
    if x<MnX then MnX:=x else if x>MxX then MxX:=x;
    if y<MnY then MnY:=y else if y>MxY then MxY:=y;
    if z<MnZ then MnZ:=z;
   end;
  end;
 end;
{$ELSE}
 var
  _Coords:pointer;
  _Count :word;
 begin
  _Coords:=Coords;
  _Count :=VerticeCount;
  asm
  les di,_Coords
  mov cx,_Count
  @next_point:

   mov bx,es:[di+4]
   add bx,CENTER_Z

   mov ax,es:[di]; cwd; mov dl,ah; mov ah,al; mov al,0
  idiv bx
   add ax,CENTER_X
   mov es:[di],ax
   add di,2

   mov ax,es:[di]; cwd; mov dl,ah; mov ah,al; mov al,0
  idiv bx
   mov bx,CENTER_Y
   sub bx,ax
   mov es:[di],bx
   add di,4

  loop @next_point
  end;
 end;
{$ENDIF}

Procedure TObject3D.DrawGouraud(AShade:byte);
 var
  i:word;
  ShadeA:integer;
 begin
 {$IFDEF FLOAT}
  if (MxX<0)or(MxY<0)or(MnZ<-CENTER_Z)or(MnX>MaxX)or(MnY>MaxY) then exit;
 {$ENDIF}
  for i:=0 to FaceCount-1 do
   if (Hidden^[i].z>=0) then with Faces^[i] do begin
    ShadeA:=AShade+Shade(Shades^[a]);
   {$IFDEF FLOAT}
    with Coords^[a] do InitZGouraud(iRound(x),iRound(y),iRound(z),ShadeA);
    with Coords^[b] do ScanZGouraud(iRound(x),iRound(y),iRound(z),AShade+Shade(Shades^[b]));
    with Coords^[c] do ScanZGouraud(iRound(x),iRound(y),iRound(z),AShade+Shade(Shades^[c]));
    with Coords^[a] do ScanZGouraud(iRound(x),iRound(y),iRound(z),ShadeA);
   {$ELSE}
    with Coords^[a] do InitZGouraud(x,y,z,ShadeA);
    with Coords^[b] do ScanZGouraud(x,y,z,AShade+Shade(Shades^[b]));
    with Coords^[c] do ScanZGouraud(x,y,z,AShade+Shade(Shades^[c]));
    with Coords^[a] do ScanZGouraud(x,y,z,ShadeA);
   {$ENDIF}
    DrawZGouraud;
   end;
 end;

Procedure TObject3D.Draw(AShade:byte);
 var
  i:word;
 begin
 {$IFDEF FLOAT}
  if (MxX<0)or(MxY<0)or(MnZ<-CENTER_Z)or(MnX>MaxX)or(MnY>MaxY) then exit;
 {$ENDIF}
  for i:=0 to FaceCount-1 do
   if (Hidden^[i].z>=0) then with Faces^[i] do begin
   {$IFDEF FLOAT}
    with Coords^[a] do InitZPoly(iRound(x),iRound(y),iRound(z));
    with Coords^[b] do ScanZPoly(iRound(x),iRound(y),iRound(z));
    with Coords^[c] do ScanZPoly(iRound(x),iRound(y),iRound(z));
    with Coords^[a] do ScanZPoly(iRound(x),iRound(y),iRound(z));
   {$ELSE}
    with Coords^[a] do InitZPoly(x,y,z);
    with Coords^[b] do ScanZPoly(x,y,z);
    with Coords^[c] do ScanZPoly(x,y,z);
    with Coords^[a] do ScanZPoly(x,y,z);
   {$ENDIF}
    DrawZPoly(AShade+Shade(Hidden^[i]));
   end;
 end;

{- Torus -----------------------------------------------------------}

Constructor TTorus.Create(AStep1,AStep2,Radius1,Radius2:word);
 begin
  Step1:=AStep1;
  Step2:=AStep2;
  Size :=Step1*Step2;
  inherited Create(Size,2*Size);
  SetSize(Radius1,Radius2);
 end;

Procedure TTorus.SetSize(Radius1,Radius2:word);
 Var
  i:word;
  a1,a2:Float;
  s1,s2:integer;
  x1,y1:Float;
 begin
  a1:=2*PI/Step1;
  a2:=2*PI/Step2;

  i:=0;
  for s1:=0 to Step1-1 do begin
   x1:=cos(s1*a1)*Radius1;
   y1:=sin(s1*a1)*Radius1;
   for s2:=0 to Step2-1 do begin
    With Vertices^[i] do begin
     X:={$IFNDEF FLOAT}Round{$ENDIF}(x1+Cos(s1*a1)*cos(s2*a2)*Radius2);
     Y:={$IFNDEF FLOAT}Round{$ENDIF}(y1+Sin(s1*a1)*cos(s2*a2)*Radius2);
     Z:={$IFNDEF FLOAT}Round{$ENDIF}(sin(s2*a2)*Radius2);
    end;
    inc(i);
   end;
  end;

  i:=0;
  for s1:=0 To Step1-1 do
   for s2:=0 To Step2-1 do begin
    with Faces^[i] do begin
     a:= s1*Step2+ s2;
     b:= s1*Step2+(s2+1) mod Step2;
     c:=(s1*Step2+ s2+Step2) mod Size;
    end;
    inc(i);
    with Faces^[i] do begin
     a:= s1*Step2+(s2+1) mod Step2;
     b:=(s1*Step2+(s2+1) mod Step2 + Step2) mod Size;
     c:=(s1*Step2+ s2+Step2) mod Size;
    end;
    inc(i);
   end;

 end;

{- Cube ------------------------------------------------------------}

Constructor TCube.Create(Width,Height,Depth:word);
 const
  size:array[1..3*8] of int=(
   -1,-1,+1,
   +1,-1,+1,
   -1,+1,+1,
   +1,+1,+1,
   -1,-1,-1,
   +1,-1,-1,
   -1,+1,-1,
   +1,+1,-1
  );
  edge:array[1..3*12] of word=(
   0,2,1, 1,2,3,
   4,5,6, 5,7,6,
   0,1,4, 4,1,5,
   6,7,2, 2,7,3,
   0,4,6, 6,2,0,
   7,5,1, 7,1,3
  );
 var
  i:integer;
 begin
  inherited Create(8,12);
  Width :=Width  div 2;
  Height:=Height div 2;
  Depth :=Depth  div 2;
  move(size,Vertices^,SizeOf(Size));
  for i:=0 to VerticeCount-1 do with Vertices^[i] do begin
   x:=x*Width;
   y:=y*Height;
   z:=z*Depth;
  end;
  move(edge,Faces^,SizeOf(edge));
 end;

{- Cylinder -------------------------------------------------------}
Constructor TCylinder.Create(Step,Height,Radius1,Radius2:word);
 var
  size:word;
  a:float;
  s:integer;
  i:integer;
 begin
  size:=2*step;
  Height:=Height div 2;
  inherited Create(Size+2,2*Size);

  with Vertices^[size] do begin x:=0; y:=0; z:=-Height end;
  with Vertices^[size+1] do begin x:=0; y:=0; z:=+Height end;

  a:=2*PI/Step;
  i:=0;
  for s:= 0 to Step-1 do begin
   With Vertices^[i] do begin
    X:=Round(Cos(s*a)*Radius1);
    Y:=Round(Sin(s*a)*Radius1);
    Z:=-Height;
   end;
   inc(i);
   With Vertices^[i] do begin
    X:=Round(Cos(s*a)*Radius2);
    Y:=Round(Sin(s*a)*Radius2);
    Z:=+Height;
   end;
   inc(i);
  end;

  i:=0;
  for s:= 0 to Step-1 do begin
   With Faces^[i] do begin
    a:=size;
    b:=(2*s+0) mod size;
    c:=(2*s+2) mod size;
   end;
   inc(i);
   With Faces^[i] do begin
    a:=size+1;
    b:=(2*s+3) mod size;
    c:=(2*s+1) mod size;
   end;
   inc(i);
  end;

  for s:= 0 to Step-1 do begin
   With Faces^[i] do begin
    a:=(2*s+0) mod size;
    b:=(2*s+1) mod size;
    c:=(2*s+2) mod size;
   end;
   inc(i);
   With Faces^[i] do begin
    a:=(2*s+3) mod size;
    b:=(2*s+2) mod size;
    c:=(2*s+1) mod size;
   end;
   inc(i);
  end;
 end;

{- Sphere ---------------------------------------------------------}
Constructor TSphere.Create(Step,Radius:word);
 var
  Size:word;
  a1:Float;
  s1:integer;
  s2:integer;
   i:integer;
  z2:float;
 begin
  a1:=PI/Step;
  Size:=Step*Step+Step;

  inherited Create(Size,2*Size);

  i:=0;
  for s1:=0 to Step do begin
   z2:=cos(s1*a1)*Radius;
   for s2:=0 to Step-1 do begin
    With Vertices^[i] do begin
     x:={$IFNDEF FLOAT}round{$ENDIF}(cos(s2*2*a1)*sin(s1*a1)*Radius);
     y:={$IFNDEF FLOAT}round{$ENDIF}(sin(s2*2*a1)*sin(s1*a1)*Radius);
     z:={$IFNDEF FLOAT}round{$ENDIF}(z2);
    end;
    inc(i);
   end;
  end;

  i:=0;
  for s1:=0 to Step do
   for s2:=0 to Step-1 do begin
    with Faces^[i] do begin
     a:= s1*Step+ s2;
     b:= s1*Step+(s2+1) mod Step;
     c:=(s1*Step+ s2+Step) mod Size;
    end;
    inc(i);
    with Faces^[i] do begin
     a:= s1*Step+(s2+1) mod Step;
     b:=(s1*Step+(s2+1) mod Step + Step) mod Size;
     c:=(s1*Step+ s2+Step) mod Size;
    end;
    inc(i);
   end;
 end;

{- GSphere ---------------------------------------------------------}
Constructor TGSphere.Create(Step,Radius:word);
 var
  Size:word;
  a1:Float;
  s1:integer;
  s2:integer;
   i:integer;
  z2:float;
  da:float;
  ds:integer;
 begin
  a1:=PI/Step;
  Size:=Step*Step+Step;

  inherited Create(Size,2*Size);

  i:=0;
  for s1:=0 to Step do begin
   z2:=cos(s1*a1)*Radius;
   if odd(s1) then da:=+a1 else da:=0;
   for s2:=0 to Step-1 do begin
    With Vertices^[i] do begin
     x:={$IFNDEF FLOAT}round{$ENDIF}(cos(s2*2*a1+da)*sin(s1*a1)*Radius);
     y:={$IFNDEF FLOAT}round{$ENDIF}(sin(s2*2*a1+da)*sin(s1*a1)*Radius);
     z:={$IFNDEF FLOAT}round{$ENDIF}(z2);
    end;
    inc(i);
   end;
  end;

  i:=0;
  for s1:=0 to Step do begin
   if odd(s1) then ds:=+1 else ds:=0;
   for s2:=0 to Step-1 do begin
    with Faces^[i] do begin
     a:= s1*Step+ s2;
     b:= s1*Step+(s2+1) mod Step;
     c:=(s1*Step+(s2+ds) mod Step+Step) mod Size;
    end;
    inc(i);
    with Faces^[i] do begin
     a:= s1*Step+(s2+1) mod Step;
     b:=(s1*Step+(s2+1+ds) mod Step+Step) mod Size;
     c:=(s1*Step+(s2+ds) mod Step+Step) mod Size;
    end;
    inc(i);
   end;
  end;
 end;

{$IFNDEF FLOAT}
var
 x:byte;
begin
 for x:=0 to 255 do begin
  sintab[x]:=Round(sin(x*PI/128)*FACTOR);
  costab[x]:=Round(cos(x*PI/128)*FACTOR);
 end;
{$ENDIF}

end.